from unittest import TestCase
from unittest.mock import Mock

from identitylib.identifiers import Identifier, IdentifierSchemes
from rest_framework.request import HttpRequest

from apigatewayauth.authentication import (
    APIGatewayAuthentication,
    APIGatewayAuthenticationDetails,
)
from apigatewayauth.patch import dummy_authentication, mock_authentication


class APIGatewayAuthTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.authn = APIGatewayAuthentication()
        self.request: HttpRequest = Mock(spec=HttpRequest)
        self.request.META = {}

    def test_dummy_returns_expected(self):
        with dummy_authentication():
            result = self.authn.authenticate(self.request)
        assert result is not None
        user, details = result
        assert user.username == str(details.principal_identifier)
        assert details.principal_identifier == Identifier("foo", IdentifierSchemes.CRSID)
        assert details.scopes == set()

    def test_mock_returns_expected(self):
        expected_details = APIGatewayAuthenticationDetails(
            principal_identifier=Identifier("Brian", IdentifierSchemes.CRSID), scopes={"read"}
        )
        with mock_authentication(expected_details):
            result = self.authn.authenticate(self.request)
        assert result is not None
        user, details = result
        assert user.username == str(expected_details.principal_identifier)
        assert details == expected_details

    def test_unauthenticated_request_returns_none(self):
        result = self.authn.authenticate(self.request)
        assert result is None



from logging import getLogger
from typing import TYPE_CHECKING, Any

from django.conf import settings
from drf_spectacular.extensions import OpenApiAuthenticationExtension
from identitylib.identifiers import Identifier
from rest_framework import authentication
from rest_framework.exceptions import AuthenticationFailed
from rest_framework.request import HttpRequest

from .id_token import InvalidIdTokenError, verify_id_token_for_api_backend
from .user import APIGatewayAuthenticationDetails, APIGatewayUser

if TYPE_CHECKING:
    # Avoids circular dependency when referenced in DEFAULT_AUTHENTICATION_CLASSES:
    # [First encountered View] -> ... -> *rest_framework.generics*.View ->
    #   Django loading of Settings, loading classes from strings ->
    #   apigatewayauth.authentication.APIGatewayAuthentication ->
    #   drf_spectacular.openapi.AutoSchema -> *rest_framework.generics*
    from drf_spectacular.openapi import AutoSchema


LOG = getLogger(__name__)


class APIGatewayAuthentication(authentication.BaseAuthentication):
    """
    An Authentication provider which interprets the headers provided by the API Gateway.

    This library expects to only be used within an application that is deployed behind and can
    only be invoked by the API Gateway, and therefore relies on the fact that the headers
    provided are authoritative.

    """

    def authenticate(self, request: HttpRequest):
        if not request.META.get("HTTP_X_API_ORG_NAME", None):
            # bail early if we look like we're not being called by the API Gateway
            return None

        try:
            # We should have "Bearer ..." in the authorization header.
            auth_header = request.headers.get("Authorization", "")
            if not auth_header.startswith("Bearer "):
                raise AuthenticationFailed("Bearer token not present")

            # Extract the id token from the API Gateway for verification.
            _, token = auth_header.split(" ")

            expected_audiences = getattr(
                settings,
                "API_GATEWAY_JWT_EXPECTED_AUDIENCE",
                [request.build_absolute_uri("/"), request.build_absolute_uri("/").rstrip("/")],
            )

            trusted_issuers = getattr(settings, "API_GATEWAY_JWT_TRUSTED_ISSUERS", None)
            expected_authorised_parties = getattr(
                settings, "API_GATEWAY_JWT_EXPECTED_AUTHORISED_PARTIES", None
            )
            try:
                verify_id_token_for_api_backend(
                    token,
                    expected_audiences,
                    certs_url=getattr(settings, "API_GATEWAY_JWT_ISSUER_CERTS_URL", None),
                    trusted_issuers=trusted_issuers,
                    expected_authorised_parties=expected_authorised_parties,
                )
            except InvalidIdTokenError as e:
                LOG.info(f"Incoming API token failed verification: {e}")
                raise AuthenticationFailed("Invalid API Gateway token") from e
        except AuthenticationFailed as e:
            if getattr(settings, "API_GATEWAY_ENFORCE_ID_TOKEN_VERIFICATION", False):
                raise e
            else:
                LOG.warning(
                    "API_GATEWAY_ENFORCE_ID_TOKEN_VERIFICATION is False. "
                    f"Allowing incoming request with invalid authentication: {e}"
                )

        if not request.META.get("HTTP_X_API_OAUTH2_USER", None):
            raise AuthenticationFailed("Could not authenticate using x-api-* headers")

        try:
            principal_identifier = Identifier.from_string(
                request.META["HTTP_X_API_OAUTH2_USER"], find_by_alias=True
            )
        except Exception:
            raise AuthenticationFailed("Invalid principal identifier")

        auth = APIGatewayAuthenticationDetails(
            principal_identifier=principal_identifier,
            scopes=set(filter(bool, request.META.get("HTTP_X_API_OAUTH2_SCOPE", "").split(" "))),
            # the following will only be populated for confidential clients
            app_id=request.META.get("HTTP_X_API_DEVELOPER_APP_ID", None),
            client_id=request.META.get("HTTP_X_API_OAUTH2_CLIENT_ID", None),
        )
        user = APIGatewayUser(auth)
        return user, auth


class APIGatewaySecuritySchemeBase(OpenApiAuthenticationExtension):
    """
    Describes the security scheme of the API Gateway authentication to allow drf spectacular
    to propagate the security requirements when using the  APIGatewayAuthentication as an
    authentication provider.
    """

    target_class = APIGatewayAuthentication
    name = APIGatewayAuthentication.__name__
    scopes: dict[str, str] = {}

    def get_security_requirement(self, auto_schema: "AutoSchema") -> dict[str, list[str]]:
        """
        Get the security requirement for a given view from the 'get_required_scopes' method on a
        view.

        """
        return {
            self.name: (
                list(auto_schema.view.get_required_scopes())
                if hasattr(auto_schema.view, "get_required_scopes")
                else []
            )
        }

    def get_security_definition(self, auto_schema: "AutoSchema") -> dict[str, Any]:
        return {
            "type": "oauth2",
            "description": "API Gateway client credentials security scheme",
            "flows": {
                "clientCredentials": {
                    "tokenUrl": "https://api.apps.cam.ac.uk/oauth2/v1/token",
                    "scopes": self.scopes,
                },
                "authorizationCode": {
                    "authorizationUrl": "https://api.apps.cam.ac.uk/oauth2/v1/auth",
                    "tokenUrl": "https://api.apps.cam.ac.uk/oauth2/v1/token",
                    "scopes": self.scopes,
                },
            },
        }



from dataclasses import dataclass
from typing import Optional, Set

from django.contrib.auth.models import AnonymousUser
from identitylib.identifiers import Identifier


@dataclass(eq=True)
class APIGatewayAuthenticationDetails:
    """
    A dataclass representing the authentication information passed from the API Gateway.

    """

    principal_identifier: Identifier
    scopes: Set[str]
    app_id: Optional[str] = None
    client_id: Optional[str] = None


class APIGatewayUser(AnonymousUser):
    """
    A Django user representing the authenticated principal. This user is not
    backed by a database object and so they can have no permissions in the
    Django sense.
    """

    def __init__(self, auth: APIGatewayAuthenticationDetails):
        super().__init__()
        self.username = self.id = self.pk = str(auth.principal_identifier)

    @property
    def is_anonymous(self):
        return False

    @property
    def is_authenticated(self):
        return True

    def __str__(self):
        return self.username



from typing import Any
from unittest import TestCase

from django.db.models import Model
from django.urls import path
from drf_spectacular.extensions import OpenApiAuthenticationExtension
from drf_spectacular.generators import SchemaGenerator
from drf_spectacular.validation import validate_schema
from rest_framework.authentication import BaseAuthentication
from rest_framework.generics import GenericAPIView
from rest_framework.permissions import IsAuthenticated
from rest_framework.routers import SimpleRouter
from rest_framework.serializers import ModelSerializer
from rest_framework.viewsets import GenericViewSet, ReadOnlyModelViewSet

from apigatewayauth.authentication import (
    APIGatewayAuthentication,
    APIGatewaySecuritySchemeBase,
)

PATHS = "paths"
COMPONENTS = "components"
SECURITY_SCHEMES = "securitySchemes"
SECURITY = "security"


class SimpleModel(Model):
    pass


class SimpleSerializer(ModelSerializer):
    class Meta:
        model = SimpleModel
        fields = "__all__"


class ScopedAuth(APIGatewayAuthentication):
    ...


class ScopedScheme(APIGatewaySecuritySchemeBase):
    target_class = ScopedAuth
    name: str = ScopedAuth.__name__
    scopes: dict[str, str] = {"scope": ""}


class UnsecuredView(GenericAPIView):
    serializer_class = SimpleSerializer
    authentication_classes: list[type[BaseAuthentication]] = []

    def get(self, request, format=None):
        ...


class SecuredView(UnsecuredView):
    authentication_classes = [APIGatewayAuthentication]
    permission_classes = [IsAuthenticated]


class ScopedView(SecuredView):
    authentication_classes = [ScopedAuth]

    def get_required_scopes(self) -> list[str]:
        return ["scope"]


class UnsecuredViewset(ReadOnlyModelViewSet):
    queryset = SimpleModel.objects.none()
    serializer_class = SimpleSerializer
    authentication_classes: list[type[BaseAuthentication]] = []


class SecuredViewset(UnsecuredViewset):
    authentication_classes = [APIGatewayAuthentication]
    permission_classes = [IsAuthenticated]


class ScopedViewset(SecuredViewset):
    authentication_classes = [ScopedAuth]

    def get_required_scopes(self) -> list[str]:
        return ["scope"]


def generate_schema(*views: GenericAPIView | type[GenericViewSet]) -> dict[str, Any]:
    patterns = []
    router = SimpleRouter()
    for view in views:
        if isinstance(view, GenericAPIView):
            patterns.append(path(type(view).__name__, view.as_view()))
        else:
            router.register(view.__name__, view, basename=view.__name__)
    patterns += router.urls
    schema = SchemaGenerator(patterns=patterns).get_schema(public=True)
    validate_schema(schema)  # make sure generated schemas are always valid
    return schema


def expected_security_scheme(scopes: dict[str, str]) -> dict[str, Any]:
    return {
        "type": "oauth2",
        "description": "API Gateway client credentials security scheme",
        "flows": {
            "clientCredentials": {
                "tokenUrl": "https://api.apps.cam.ac.uk/oauth2/v1/token",
                "scopes": scopes,
            },
            "authorizationCode": {
                "authorizationUrl": "https://api.apps.cam.ac.uk/oauth2/v1/auth",
                "tokenUrl": "https://api.apps.cam.ac.uk/oauth2/v1/token",
                "scopes": scopes,
            },
        },
    }


UNSCOPED_SCHEME = expected_security_scheme({})

SCOPED_SCHEME = expected_security_scheme({"scope": ""})

UNSCOPED_SECURITY: list[dict[type[BaseAuthentication], list[str]]] = [
    {APIGatewayAuthentication.__name__: []}
]
SCOPED_SECURITY: list[dict[type[BaseAuthentication], list[str]]] = [
    {ScopedAuth.__name__: ["scope"]}
]


class APIGatewayAuthTestCase(TestCase):
    def assert_schema_contains_security_scheme(
        self,
        schema: dict[str, Any],
        expected_scheme: dict[type[APIGatewayAuthentication], dict[str, Any]],
        security_scheme: type[BaseAuthentication] = APIGatewayAuthentication,
    ):
        self.assertIn(COMPONENTS, schema)
        self.assertIn(SECURITY_SCHEMES, schema[COMPONENTS])
        self.assertEqual(
            expected_scheme, schema[COMPONENTS][SECURITY_SCHEMES][security_scheme.__name__]
        )

    def assert_endpoints_include_security(
        self,
        schema: dict[str, Any],
        endpoints: list[str],
        expected_scopes: list[dict[str, list[str]]],
    ):
        for endpoint in endpoints:
            self.assertIn(endpoint, schema[PATHS])
            for operation in schema[PATHS][endpoint]:
                self.assertEqual(expected_scopes, schema[PATHS][endpoint][operation][SECURITY])

    def assert_all_endpoints_checked(self, schema: dict[str, Any], *tested_endpoints: list[str]):
        expected_paths = set(a for endpoint in tested_endpoints for a in endpoint)
        self.assertEqual(expected_paths, set(schema[PATHS]))

    def test_unauthenticated_view_no_security(self):
        """
        When:
        - DEFAULT_AUTHENTICATION_CLASSES is not defined
        - View does not define authentication_classes
        - APIGatewaySecuritySchemeBase is registered
        Then:
        - Generated schema for view does not include securityScheme
        - Generated schema for view does not include security on
          operations defined by view.
        """
        schema = generate_schema(UnsecuredView())

        self.assertNotIn(SECURITY_SCHEMES, schema[COMPONENTS])
        for endpoint in schema[PATHS]:
            for operation in schema[PATHS][endpoint]:
                self.assertNotIn(SECURITY, operation)

    def test_unauthenticated_viewset_no_security(self):
        """
        When:
        - DEFAULT_AUTHENTICATION_CLASSES is not defined
        - Viewset does not define authentication_classes
        - APIGatewaySecuritySchemeBase is registered
        Then:
        - Generated schema for viewset does not include securityScheme
        - Generated schema for viewset does not include security on
          operations defined by viewset.
        """
        schema = generate_schema(UnsecuredViewset)

        self.assertNotIn(SECURITY_SCHEMES, schema[COMPONENTS])
        for endpoint in schema[PATHS]:
            for operation in schema[PATHS][endpoint]:
                self.assertNotIn(SECURITY, operation)

    def test_authenticated_view_describes_security(self):
        """
        When:
        - View authentication_classes includes APIGatewayAuthentication
        - APIGatewaySecuritySchemeBase is registered
        Then:
        - Generated schema for view includes securityScheme from
          APIGatewaySecuritySchemeBase
        - Generated schema for view includes security on each operation
          from view
        """
        schema = generate_schema(SecuredView())

        secured_endpoints = ["/SecuredView"]

        self.assert_schema_contains_security_scheme(schema, UNSCOPED_SCHEME)
        self.assert_endpoints_include_security(schema, secured_endpoints, UNSCOPED_SECURITY)
        self.assert_all_endpoints_checked(schema, secured_endpoints)

    def test_authenticated_viewset_describes_security(self):
        """
        When:
        - Viewset authentication_classes includes APIGatewayAuthentication
        - APIGatewaySecuritySchemeBase is registered
        Then:
        - Generated schema for viewset includes securityScheme from
          APIGatewaySecuritySchemeBase
        - Generated schema for viewset includes security on each operation
          from view
        """
        schema = generate_schema(SecuredViewset)
        secured_endpoints = ["/SecuredViewset/", "/SecuredViewset/{id}/"]

        self.assert_schema_contains_security_scheme(schema, UNSCOPED_SCHEME)
        self.assert_endpoints_include_security(schema, secured_endpoints, UNSCOPED_SECURITY)
        self.assert_all_endpoints_checked(schema, secured_endpoints)

    def test_authenticated_view_security_generated_with_scopes(self):
        """
        When:
        - View authentication_classes includes APIGatewayAuthentication
        - APIGatewaySecuritySchemeBase is registered
        - View defines 'get_required_scopes' method
        Then:
        - Generated schema for view includes securityScheme from
          APIGatewaySecuritySchemeBase
        - Generated schema for view includes security on each operation
          from view
        - Generated security on each operation of view includes scopes defined
          on view
        """
        schema = generate_schema(ScopedView())
        secured_endpoints = ["/ScopedView"]

        self.assert_schema_contains_security_scheme(schema, SCOPED_SCHEME, ScopedAuth)
        self.assert_endpoints_include_security(schema, secured_endpoints, SCOPED_SECURITY)
        self.assert_all_endpoints_checked(schema, secured_endpoints)

    def test_authenticated_viewset_security_generated_with_scopes(self):
        """
        When:
        - Viewset authentication_classes includes APIGatewayAuthentication
        - APIGatewaySecuritySchemeBase is registered
        - Viewset defines 'get_required_scopes' method
        Then:
        - Generated schema for viewset includes securityScheme from
          APIGatewaySecuritySchemeBase
        - Generated schema for viewset includes security on each operation
          from viewset
        - Generated security on each operation of view includes scopes defined
          on viewset
        """
        schema = generate_schema(ScopedViewset)
        secured_endpoints = ["/ScopedViewset/", "/ScopedViewset/{id}/"]

        self.assert_schema_contains_security_scheme(schema, SCOPED_SCHEME, ScopedAuth)
        self.assert_endpoints_include_security(schema, secured_endpoints, SCOPED_SECURITY)
        self.assert_all_endpoints_checked(schema, secured_endpoints)

    def test_authenticated_view_security_generated_with_specific_scopes(self):
        """
        When:
        - APIGatewaySecuritySchemeBase is registered
        - Two views include APIGatewayAuthentication in authentication_classes
        - View_a defines 'get_required_scopes' method
        - View_b does not define 'get_required_scopes' method
        Then:
        - Generated schema for views includes securityScheme from
          APIGatewaySecuritySchemeBase
        - Generated schema for views includes security on each operation
          from both views
        - Generated security on each operation of view_a includes scopes defined
          on view
        - Generated security on each operation of view_b does not include scopes,
          as none defined on view
        """
        schema = generate_schema(SecuredView(), ScopedView())

        scoped_endpoints = ["/ScopedView"]
        unscoped_endpoints = ["/SecuredView"]

        self.assert_schema_contains_security_scheme(schema, UNSCOPED_SCHEME)
        self.assert_schema_contains_security_scheme(schema, SCOPED_SCHEME, ScopedAuth)
        self.assert_endpoints_include_security(schema, scoped_endpoints, SCOPED_SECURITY)
        self.assert_endpoints_include_security(schema, unscoped_endpoints, UNSCOPED_SECURITY)
        self.assert_all_endpoints_checked(schema, unscoped_endpoints, scoped_endpoints)

    def test_authenticated_viewset_security_generated_with_specific_scopes(self):
        """
        When:
        - APIGatewaySecuritySchemeBase is registered
        - Two viewsets include APIGatewayAuthentication in authentication_classes
        - Viewset_a defines 'get_required_scopes' method
        - Viewset_b does not define 'get_required_scopes' method
        Then:
        - Generated schema for viewsets includes securityScheme from
          APIGatewaySecuritySchemeBase
        - Generated schema for viewsets includes security on each operation
          from both viewsets
        - Generated security on each operation of view_a includes scopes defined
          on viewset
        - Generated security on each operation of view_b does not include scopes,
          as none defined on viewset
        """
        schema = generate_schema(SecuredViewset, ScopedViewset)

        scoped_endpoints = ["/ScopedViewset/", "/ScopedViewset/{id}/"]
        unscoped_endpoints = ["/SecuredViewset/", "/SecuredViewset/{id}/"]

        self.assert_schema_contains_security_scheme(schema, UNSCOPED_SCHEME)
        self.assert_endpoints_include_security(schema, scoped_endpoints, SCOPED_SECURITY)
        self.assert_endpoints_include_security(schema, unscoped_endpoints, UNSCOPED_SECURITY)
        self.assert_all_endpoints_checked(schema, unscoped_endpoints, scoped_endpoints)

    def test_prioritised_redefinition_includes_scopes_for_view(self):
        """
        When:
        - APIGatewaySecuritySchemeBase is registered
        - ScopedScheme(OpenApiAuthenticationExtension) is registered with
          `target_class = APIGatewayAuthentication` and priority > 0
        - View authentication_classes includes APIGatewayAuthentication
        Then:
        - Generated schema for view includes securityScheme from ScopedScheme
        - Generated schema for view includes security on each operation
          from view
        """

        class ScopedScheme(APIGatewaySecuritySchemeBase):
            scopes: dict[str, str] = {"foo": ""}
            priority: int = 1

        try:
            schema = generate_schema(SecuredView())
            secured_endpoints = ["/SecuredView"]

            self.assert_schema_contains_security_scheme(
                schema, expected_security_scheme({"foo": ""})
            )
            self.assert_endpoints_include_security(schema, secured_endpoints, UNSCOPED_SECURITY)
            self.assert_all_endpoints_checked(schema, secured_endpoints)

        finally:
            OpenApiAuthenticationExtension._registry.remove(ScopedScheme)
            del ScopedScheme

    def test_prioritised_redefinition_includes_scopes_for_viewset(self):
        """
        When:
        - APIGatewaySecuritySchemeBase is registered
        - ScopedScheme(APIGatewaySecuritySchemeBase) is registered with
          `target_class = APIGatewayAuthentication` and priority > 0
        - Viewset authentication_classes includes APIGatewayAuthentication
        Then:
        - Generated schema for viewset includes securityScheme from
          other ScopedScheme
        - Generated schema for viewset includes security on each operation
          from viewset
        """

        class ScopedScheme(APIGatewaySecuritySchemeBase):
            scopes: dict[str, str] = {"foo": ""}
            priority: int = 1

        try:
            schema = generate_schema(SecuredViewset)
            secured_endpoints = ["/SecuredViewset/", "/SecuredViewset/{id}/"]

            self.assert_schema_contains_security_scheme(
                schema, expected_security_scheme({"foo": ""})
            )
            self.assert_endpoints_include_security(schema, secured_endpoints, UNSCOPED_SECURITY)
            self.assert_all_endpoints_checked(schema, secured_endpoints)

        finally:
            OpenApiAuthenticationExtension._registry.remove(ScopedScheme)
            del ScopedScheme

    def test_views_independent(self):
        """
        When:
        - BaseAuthentication_a is in the authentication_classes of view_a,
          and described by OpenApiAuthenticationExtension_a, which is registered
        - BaseAuthentication_b is in the authentication_classes of view_b,
          and described by OpenApiAuthenticationExtension_b, which is registered
        Then:
        - Generated schema for views includes securitySchemes described by
          OpenApiAuthenticationExtension_a and OpenApiAuthenticationExtension_b
        - Generated schema for view_a and view_b includes security on each operation
          from both views
        - Generated security on each operation of view_a references securityScheme
          defined on OpenApiAuthenticationExtension_a
        - Generated security on each operation of view_b references securityScheme
          defined on OpenApiAuthenticationExtension_b
        """
        schema = generate_schema(ScopedView(), SecuredView())

        scoped_endpoints = ["/ScopedView"]
        unscoped_endpoints = ["/SecuredView"]

        self.assert_schema_contains_security_scheme(schema, SCOPED_SCHEME, ScopedAuth)
        self.assert_schema_contains_security_scheme(schema, UNSCOPED_SCHEME)
        self.assert_endpoints_include_security(schema, scoped_endpoints, SCOPED_SECURITY)
        self.assert_endpoints_include_security(schema, unscoped_endpoints, UNSCOPED_SECURITY)
        self.assert_all_endpoints_checked(schema, scoped_endpoints, unscoped_endpoints)

    def test_viewsets_independent(self):
        """
        When:
        - BaseAuthentication_a is in the authentication_classes of viewset_a,
          and described by OpenApiAuthenticationExtension_a, which is registered
        - BaseAuthentication_b is in the authentication_classes of viewset_b,
          and described by OpenApiAuthenticationExtension_b, which is registered
        Then:
        - Generated schema for viewsets includes securitySchemes described by
          OpenApiAuthenticationExtension_a and OpenApiAuthenticationExtension_b
        - Generated schema for viewset_a and viewset_b includes security on each
          operation from both viewsets
        - Generated security on each operation of viewset_a references securityScheme
          defined on OpenApiAuthenticationExtension_a
        - Generated security on each operation of viewset_b references securityScheme
          defined on OpenApiAuthenticationExtension_b
        """

        schema = generate_schema(SecuredViewset, ScopedViewset)

        scoped_endpoints = ["/SecuredViewset/", "/SecuredViewset/{id}/"]
        unscoped_endpoints = ["/ScopedViewset/", "/ScopedViewset/{id}/"]

        self.assert_schema_contains_security_scheme(schema, SCOPED_SCHEME, ScopedAuth)
        self.assert_schema_contains_security_scheme(schema, UNSCOPED_SCHEME)
        self.assert_endpoints_include_security(schema, unscoped_endpoints, SCOPED_SECURITY)
        self.assert_endpoints_include_security(schema, scoped_endpoints, UNSCOPED_SECURITY)
        self.assert_all_endpoints_checked(schema, unscoped_endpoints, scoped_endpoints)



from typing import ContextManager
from unittest import mock

from identitylib.identifiers import Identifier, IdentifierSchemes

from apigatewayauth.authentication import APIGatewayAuthenticationDetails
from apigatewayauth.user import APIGatewayUser


def dummy_authentication() -> ContextManager:
    """
    Preconfigured patch to allow tests to bypass authentication,
    returning a non-real user.
    """
    return mock_authentication(
        APIGatewayAuthenticationDetails(
            principal_identifier=Identifier("foo", IdentifierSchemes.CRSID), scopes=set()
        )
    )


def mock_authentication(details: APIGatewayAuthenticationDetails) -> ContextManager:
    """
    Configurable patch to allow tests to bypass authentication,
    returning a user with expected details, e.g to test authorization.
    """
    return mock.patch(
        "apigatewayauth.authentication.APIGatewayAuthentication.authenticate",
        mock.Mock(return_value=(APIGatewayUser(details), details)),
    )



from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies: list[str] = []

    operations = [
        migrations.CreateModel(
            name="TestModel",
            fields=[
                (
                    "name",
                    models.TextField(primary_key=True, serialize=False, verbose_name="Name"),
                ),
                ("isAdmin", models.BooleanField(verbose_name="Is Admin")),
                (
                    "principal_identifier",
                    models.TextField(verbose_name="Principal identifier"),
                ),
            ],
        ),
    ]



from unittest.mock import MagicMock, patch

from django.test import TestCase
from identitylib.identifiers import Identifier, IdentifierSchemes
from rest_framework.test import APIRequestFactory
from ucamlookup.ibisclient import IbisException

from apigatewayauth.authentication import APIGatewayAuthenticationDetails
from apigatewayauth.permissions import (
    Disallowed,
    HasAnyScope,
    IsResourceOwningPrincipal,
    SpecifiedPermission,
    get_permissions_for_request,
)

from .mocks.mock_ibis import MockIbisGroup
from .mocks.models import TestModel
from .mocks.permissions_spec import override_permission_spec


class PermissionsTestCase(TestCase):
    def compatAssertQuerySetEqual(self, *args, **kwargs):
        """
        Backwards compatibility function as TestCase class refactored assertQuerysetEqual to
        assertQuerySetEqual in Django v5.
        """
        if hasattr(self, "assertQuerysetEqual"):
            return self.assertQuerysetEqual(*args, **kwargs)
        return self.assertQuerySetEqual(*args, **kwargs)

    def setUp(self):
        super().setUp()

        self.view = MagicMock()
        self.object = MagicMock()

        self.client_auth_details = APIGatewayAuthenticationDetails(
            Identifier("abc", IdentifierSchemes.CRSID), set()
        )

    def request_with_auth(self, auth):
        request = APIRequestFactory().get("/")
        setattr(request, "auth", auth)
        return request

    def test_disallowed_permission_always_disallows(self):
        permission = Disallowed()

        for auth in [None, self.client_auth_details]:
            self.assertFalse(permission.has_permission(self.request_with_auth(auth), self.view))
            self.assertFalse(
                permission.has_object_permission(
                    self.request_with_auth(auth), self.view, self.object
                )
            )

    def test_is_resource_owning_principal_returns_true_when_authenticated(self):
        permission = IsResourceOwningPrincipal()

        self.assertFalse(permission.has_permission(self.request_with_auth(None), self.view))
        self.assertFalse(
            permission.has_object_permission(self.request_with_auth(None), self.view, self.object)
        )

        public_client_request = self.request_with_auth(self.client_auth_details)
        self.assertTrue(permission.has_permission(public_client_request, self.view))
        # the request should be flagged as needing limiting
        self.assertTrue(
            getattr(public_client_request, "should_limit_to_resource_owning_principal")
        )

    def test_is_resource_owning_principal_users_has_object_permission(self):
        permission = IsResourceOwningPrincipal()

        # should be false as the object passed in doesn't have a 'is_owned_by' property
        self.assertFalse(
            permission.has_object_permission(
                self.request_with_auth(self.client_auth_details), self.view, {}
            )
        )

        mock_object = MagicMock()
        mock_object.is_owned_by.return_value = False

        self.assertFalse(
            permission.has_object_permission(
                self.request_with_auth(self.client_auth_details), self.view, mock_object
            )
        )
        mock_object.is_owned_by.assert_called_with(self.client_auth_details.principal_identifier)

        mock_object.is_owned_by.return_value = True
        self.assertTrue(
            permission.has_object_permission(
                self.request_with_auth(self.client_auth_details), self.view, mock_object
            )
        )

    def test_get_queryset_for_principal_bails_for_unhandled_request(self):
        # bails early if request doesn't require filtering
        self.compatAssertQuerySetEqual(
            IsResourceOwningPrincipal.get_queryset_for_principal(
                self.request_with_auth(self.client_auth_details), TestModel
            ),
            TestModel.objects.all(),
        )

    def test_get_queryset_for_principal_returns_filtered_queryset(self):
        permission = IsResourceOwningPrincipal()

        request = self.request_with_auth(self.client_auth_details)

        # pass through `has_permission` to mark `should_limit_to_resource_owning_principal`
        # on the requests
        self.assertTrue(permission.has_permission(request, self.view))

        self.compatAssertQuerySetEqual(
            IsResourceOwningPrincipal.get_queryset_for_principal(request, TestModel),
            TestModel.get_queryset_for_principal(self.client_auth_details.principal_identifier),
        )

    def test_get_queryset_for_principal_returns_no_objects_if_no_request_auth(self):
        permission = IsResourceOwningPrincipal()

        request = APIRequestFactory().get("/")

        # `has_permission` should return false as we have no auth details on the request
        self.assertFalse(permission.has_permission(request, self.view))

        # Set should_limit_to_resource_owning_principal manually, which allows us to hit the
        # condition of having no principal whilst trying to get the queryset for a principal.
        # This should never really happen, but we handle the condition in the permission class
        # so it's worth a test.
        setattr(request, "should_limit_to_resource_owning_principal", True)

        # If we get the queryset for the principal we should get objects.none() as we do not have
        # a principal.
        self.compatAssertQuerySetEqual(
            IsResourceOwningPrincipal.get_queryset_for_principal(request, TestModel),
            TestModel.objects.none(),
        )

    def test_get_queryset_for_principal_throws_on_an_unknown_model(self):
        permission = IsResourceOwningPrincipal()

        request = self.request_with_auth(self.client_auth_details)
        self.assertTrue(permission.has_permission(request, self.view))

        # using get_queryset_for_principal without a model defining get_queryset_for_principal
        # should cause a value error with a desciption message
        with self.assertRaisesRegex(
            ValueError, "{} does not implement get_queryset_for_principal"
        ):
            IsResourceOwningPrincipal.get_queryset_for_principal(request, {})

    def test_has_any_scope(self):
        permission = HasAnyScope("a", "b.readonly")()

        for auth in [None, self.client_auth_details]:
            self.assertFalse(permission.has_permission(self.request_with_auth(auth), self.view))
            self.assertFalse(
                permission.has_object_permission(
                    self.request_with_auth(auth), self.view, self.object
                )
            )

        self.assertTrue(
            permission.has_permission(
                self.request_with_auth(
                    APIGatewayAuthenticationDetails(
                        Identifier("abc", IdentifierSchemes.CRSID), set(["b.readonly"])
                    )
                ),
                self.view,
            )
        )

        self.assertFalse(
            permission.has_object_permission(
                self.request_with_auth(
                    APIGatewayAuthenticationDetails(
                        Identifier("abc", IdentifierSchemes.CRSID), set(["abcd"])
                    )
                ),
                self.view,
                self.object,
            )
        )

    @patch("apigatewayauth.permissions.get_principals_with_permission")
    def test_has_specified_permission_return_false_if_not_matching(self, get_identities_mock):
        get_identities_mock.return_value = set(
            [
                Identifier("1000", IdentifierSchemes.STAFF_NUMBER),
                Identifier("abc123", IdentifierSchemes.CRSID),
            ]
        )
        permission = SpecifiedPermission("READ_BOOKS")()

        self.assertFalse(
            permission.has_permission(self.request_with_auth(self.client_auth_details), self.view)
        )
        self.assertFalse(
            permission.has_object_permission(
                self.request_with_auth(self.client_auth_details), self.view, {}
            )
        )
        get_identities_mock.assert_called_with("READ_BOOKS")

    @patch("apigatewayauth.permissions.get_principals_with_permission")
    def test_has_specified_permission_return_true_if_matching(self, get_identities_mock):
        get_identities_mock.return_value = set(
            [
                Identifier("1000", IdentifierSchemes.STAFF_NUMBER),
                self.client_auth_details.principal_identifier,
            ]
        )
        permission = SpecifiedPermission("READ_BOOKS")()

        self.assertTrue(
            permission.has_permission(self.request_with_auth(self.client_auth_details), self.view)
        )
        self.assertTrue(
            permission.has_object_permission(
                self.request_with_auth(self.client_auth_details), self.view, {}
            )
        )
        get_identities_mock.assert_called_with("READ_BOOKS")

    @patch("apigatewayauth.permissions.get_principals_with_permission")
    @patch("apigatewayauth.permissions.get_groups_with_permission")
    @patch("apigatewayauth.permissions.PersonMethods.getGroups")
    def test_does_not_query_lookup_groups_for_non_crsid_principal(
        self,
        get_groups_mock,
        get_groups_with_permission_mock,
        get_identities_mock,
    ):
        get_identities_mock.return_value = set(
            [
                Identifier("abc123", IdentifierSchemes.CRSID),
                Identifier("abc333", IdentifierSchemes.CRSID),
            ]
        )
        get_groups_with_permission_mock.return_value = set(
            [
                Identifier("1000", IdentifierSchemes.LOOKUP_GROUP),
            ]
        )
        permission = SpecifiedPermission("READ_MAGAZINES")()

        auth_details = APIGatewayAuthenticationDetails(
            Identifier("1", IdentifierSchemes.STAFF_NUMBER), set([])
        )

        self.assertFalse(
            permission.has_permission(self.request_with_auth(auth_details), self.view)
        )
        self.assertFalse(
            permission.has_object_permission(self.request_with_auth(auth_details), self.view, {})
        )
        get_identities_mock.assert_called_with("READ_MAGAZINES")

        # should not have queried for groups as our principal does not have a crsid identifier
        get_groups_mock.assert_not_called()

    @patch("apigatewayauth.permissions.get_principals_with_permission")
    @patch("apigatewayauth.permissions.get_groups_with_permission")
    @patch("apigatewayauth.permissions.PersonMethods.getGroups")
    def test_does_not_query_lookup_groups_if_no_lookup_groups_in_permissions_spec(
        self,
        get_groups_mock,
        get_groups_with_permission_mock,
        get_identities_mock,
    ):
        get_identities_mock.return_value = set(
            [
                Identifier("1000", IdentifierSchemes.STAFF_NUMBER),
                Identifier("bb123", IdentifierSchemes.CRSID),
            ]
        )
        get_groups_with_permission_mock.return_value = set()
        permission = SpecifiedPermission("READ_PAMPHLETS")()

        self.assertFalse(
            permission.has_permission(self.request_with_auth(self.client_auth_details), self.view)
        )
        get_identities_mock.assert_called_with("READ_PAMPHLETS")

        # should not have queried for groups as we do not have a lookup group in our
        # allowed identities list
        get_groups_mock.assert_not_called()

    @patch("apigatewayauth.permissions.get_principals_with_permission")
    @patch("apigatewayauth.permissions.get_groups_with_permission")
    @patch("apigatewayauth.permissions.PersonMethods.getGroups")
    def test_will_return_true_if_principal_is_in_specified_lookup_group(
        self,
        get_groups_mock,
        get_groups_with_permission_mock,
        get_identities_mock,
    ):
        get_identities_mock.return_value = set([Identifier("cc123", IdentifierSchemes.CRSID)])
        get_groups_with_permission_mock.return_value = set(
            [Identifier("1001", IdentifierSchemes.LOOKUP_GROUP)]
        )
        permission = SpecifiedPermission("READ_THE_NEWS")()
        get_groups_mock.return_value = [MockIbisGroup("1001"), MockIbisGroup("1002")]

        self.assertTrue(
            permission.has_permission(self.request_with_auth(self.client_auth_details), self.view)
        )
        get_identities_mock.assert_called_with("READ_THE_NEWS")

        # should have queried Lookup for the principal's groups
        get_groups_mock.assert_called_with(
            scheme="crsid",
            identifier=self.client_auth_details.principal_identifier.value,
        )

        # querying again returns true without going to Lookup due to caching the result
        get_groups_mock.reset_mock()
        self.assertTrue(
            permission.has_permission(self.request_with_auth(self.client_auth_details), self.view)
        )
        get_groups_mock.assert_not_called()

    @patch("apigatewayauth.permissions.get_principals_with_permission")
    @patch("apigatewayauth.permissions.get_groups_with_permission")
    @patch("apigatewayauth.permissions.PersonMethods.getGroups")
    def test_will_return_false_if_principal_is_not_in_specified_lookup_group(
        self,
        get_groups_mock,
        get_groups_with_permission_mock,
        get_identities_mock,
    ):
        get_identities_mock.return_value = set([Identifier("dd123", IdentifierSchemes.CRSID)])
        get_groups_with_permission_mock.return_value = set(
            [Identifier("10101", IdentifierSchemes.LOOKUP_GROUP)]
        )
        permission = SpecifiedPermission("READ_THE_ROOM")()
        get_groups_mock.return_value = [MockIbisGroup("1001"), MockIbisGroup("1002")]

        self.assertFalse(
            permission.has_permission(self.request_with_auth(self.client_auth_details), self.view)
        )
        get_identities_mock.assert_called_with("READ_THE_ROOM")

        # should have queried Lookup for the principal's groups
        get_groups_mock.assert_called_with(
            scheme="crsid",
            identifier=self.client_auth_details.principal_identifier.value,
        )

    @patch("apigatewayauth.permissions.get_principals_with_permission")
    @patch("apigatewayauth.permissions.get_groups_with_permission")
    @patch("apigatewayauth.permissions.PersonMethods.getGroups")
    def test_will_return_false_if_unable_to_read_lookup_groups(
        self,
        get_groups_mock,
        get_groups_with_permission_mock,
        get_identities_mock,
    ):
        get_identities_mock.return_value = set(
            [
                Identifier("ee123", IdentifierSchemes.CRSID),
            ]
        )
        get_groups_with_permission_mock.return_value = set(
            [
                Identifier("2000", IdentifierSchemes.LOOKUP_GROUP),
            ]
        )
        permission = SpecifiedPermission("READ_THE_FUTURE")()

        def throw_ibis_error(*args, **kwargs):
            error = ValueError("failed")
            setattr(error, "message", "failed")  # ibis exception needs a 'message' attribute
            raise IbisException(error)

        get_groups_mock.side_effect = throw_ibis_error

        self.assertFalse(
            permission.has_permission(self.request_with_auth(self.client_auth_details), self.view)
        )
        get_identities_mock.assert_called_with("READ_THE_FUTURE")

        # should have queried Lookup for the principal's groups
        get_groups_mock.assert_called_with(
            scheme="crsid",
            identifier=self.client_auth_details.principal_identifier.value,
        )

    @override_permission_spec(
        {
            "DATA_READER": {
                "principals": {
                    str(Identifier("abc44", IdentifierSchemes.CRSID)),
                    str(Identifier("abc55", IdentifierSchemes.CRSID)),
                }
            },
            "DATA_OWNER": {"principals": {str(Identifier("abc44", IdentifierSchemes.CRSID))}},
        }
    )
    def test_get_permissions_for_request(self):
        """
        Test that `get_permissions_for_request` returns correctly based on the user authenticated
        on the request.

        """

        mock_request = APIRequestFactory().get("/")
        setattr(mock_request, "auth", None)

        # expect an empty list to be returned if no auth details provided
        self.assertListEqual(get_permissions_for_request(mock_request), [])

        # if we have a principal not in our permissions spec we should get an empty list
        mock_request.auth = APIGatewayAuthenticationDetails(
            Identifier("abc123", IdentifierSchemes.CRSID), []
        )
        self.assertListEqual(get_permissions_for_request(mock_request), [])

        # abc55 should be shown as a `DATA_READER`
        mock_request.auth = APIGatewayAuthenticationDetails(
            Identifier("abc55", IdentifierSchemes.CRSID), []
        )
        self.assertListEqual(get_permissions_for_request(mock_request), ["DATA_READER"])

        # abc44 should be shown as a `DATA_READER` and `DATA_WRITER`
        mock_request.auth = APIGatewayAuthenticationDetails(
            Identifier("abc44", IdentifierSchemes.CRSID), []
        )
        self.assertListEqual(
            get_permissions_for_request(mock_request), ["DATA_OWNER", "DATA_READER"]
        )
